using System;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Collections;
using System.ComponentModel;
using System.Windows.Forms;
using System.Threading;

namespace Graphing
{
	public struct Pair
	{
		private double x, y;
		public Pair(double x, double y)
		{
			this.x = x;
			this.y = y;
		}
		public double X
		{
			get
			{
				return x;
			}
			set
			{
				x = value;
			}
		}
		public double Y
		{
			get
			{
				return y;
			}
			set
			{
				y = value;
			}
		}
	}

	/// <summary>
	/// A window with graphing functionality.
	/// </summary>
	public class GraphWindow : System.Windows.Forms.Form
	{	
		private GraphWindowThread t;
		private int radius;
		private System.ComponentModel.Container components = null;
		private ArrayList data = new ArrayList();

		public Pair[] Data
		{
			set
			{
				lock(this)
				{
					data = new ArrayList(value);
				}
				Invalidate();
			}
		}

		public void Append(double x, double y)
		{
			lock(this)
			{
				data.Add(new Pair(x, y));
			}
			Invalidate();
		}

		private class GraphWindowThread
		{
			private GraphWindow window;
			private bool isReady = false;
			private string title;
			private int radius;
			private Thread t;
			public GraphWindowThread(string title)
				: this(title, 0)
			{}
			public GraphWindowThread(string title, int radius)
			{
				this.title = title;
				this.radius = radius;
				t = new Thread(new ThreadStart(start));
				t.Start();
			}
			private void start()
			{
				window = new GraphWindow(title, radius);
				isReady = true;
				while(true)
				{
					try
					{
						Application.Run(window);
					}
					catch(ThreadAbortException)
					{
						Application.ExitThread();
					}
				}
			}
			public GraphWindow Window
			{
				get
				{
					while(!isReady || !window.Created)
					{
						Thread.Sleep(10);
					}
					return window;
				}
			}
			public void SendStop()
			{
				t.Abort();
			}
		}

		public static GraphWindow Create(string title)
		{
			return Create(title, 0);
		}

		public static GraphWindow Create(string title, int radius)
		{
			GraphWindowThread t = new GraphWindowThread(title, radius);
			t.Window.t = t;
			return t.Window;
		}

		public static GraphWindow[] Create(int number)
		{
			return Create(number, 0);
		}

		public static GraphWindow[] Create(int number, int radius)
		{
			GraphWindow[] windowList = new GraphWindow[number];

			Rectangle r = SystemInformation.WorkingArea;
			int rows = (int)Math.Ceiling(Math.Sqrt(number));
			int columns = (int)Math.Ceiling(number/(double)rows);
			int width = r.Width/columns;
			int height = r.Height/rows;

			for(int i = 0; i < number; i++)
			{
				int column = i % columns;
				int row = i / columns;

				windowList[i] = GraphWindow.Create(i.ToString(), radius);
				windowList[i].Location = new Point(r.X + width*column, r.Y + height*row);
				windowList[i].Size = new Size(width, height);
			}

			return windowList;
		}

		private GraphWindow(string title, int radius)
		{
			SetStyle(ControlStyles.DoubleBuffer | ControlStyles.UserPaint | ControlStyles.AllPaintingInWmPaint, true);
			InitializeComponent();
			Text = title;
			this.radius = radius;
		}

		public void InsertFront(double x, double y)
		{
			lock(this)
			{
				data.Insert(0, new Pair(x, y));
			}
			Invalidate();
		}

		protected override void Dispose( bool disposing )
		{
			if( disposing )
			{
				if(components != null)
				{
					components.Dispose();
				}
			}
			base.Dispose( disposing );
			t.SendStop();
		}

		#region Windows Form Designer generated code
		/// <summary>
		/// Erforderliche Methode fr die Designeruntersttzung. 
		/// Der Inhalt der Methode darf nicht mit dem Code-Editor gendert werden.
		/// </summary>
		private void InitializeComponent()
		{
			// 
			// GraphWindow
			// 
			this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
			this.BackColor = System.Drawing.SystemColors.Window;
			this.ClientSize = new System.Drawing.Size(648, 342);
			this.Name = "GraphWindow";
			this.Text = "GraphWindow";

		}
		#endregion

		protected override void OnPaint(System.Windows.Forms.PaintEventArgs e)
		{
			Pair[] d;

			lock(this)
			{
				if(data == null || data.Count < 2)
				{
					return;
				}

				d = (Pair[])data.ToArray(typeof(Pair));
			}

			int w = ClientSize.Width;
			int h = ClientSize.Height;

			double xMin = 0.0; // always include the origin
			double xMax = 0.0;
			double yMin = 0.0;
			double yMax = 0.0;

			foreach(Pair p in d)
			{
				if(xMin > p.X)
				{
					xMin = p.X;
				}
				if(xMax < p.X)
				{
					xMax = p.X;
				}
				if(yMin > p.Y)
				{
					yMin = p.Y;
				}
				if(yMax < p.Y)
				{
					yMax = p.Y;
				}
			}

			double xRange = xMax - xMin;
			if(xRange < 100.0*Double.Epsilon)
			{
				xRange = 1.0;
			}
			double xWindowMin = xMin - 0.1*xRange;
			double xWindowMax = xMax + 0.1*xRange;
			double xWindowRange = 1.2*xRange;

			double yRange = yMax - yMin;
			if(yRange < 100.0*Double.Epsilon)
			{
				yRange = 1.0;
			}
			double yWindowMin = yMin - 0.1*yRange;
			double yWindowMax = yMax + 0.1*yRange;
			double yWindowRange = 1.2*yRange;

			e.Graphics.SmoothingMode = SmoothingMode.AntiAlias;

			try
			{
				//draw axes
				e.Graphics.DrawLine(Pens.Black,
					(float)( w*(-xWindowMin)/xWindowRange ), 0,
					(float)( w*(-xWindowMin)/xWindowRange ), h);
				e.Graphics.DrawLine(Pens.Black,
					0, (float)( h*(1.0-(-yWindowMin)/yWindowRange) ),
					w, (float)( h*(1.0-(-yWindowMin)/yWindowRange) ));

				//determine the scale of the axis divisions
				double xScale = Math.Pow(10.0, Math.Floor(Math.Log10(xRange/2.0)));
				double yScale = Math.Pow(10.0, Math.Floor(Math.Log10(yRange/2.0)));

				Font f = new Font("Arial Narrow", 8.0f);
				
				//draw divisions along x axis
				for(int i = (int)Math.Ceiling(xMin/xScale); i <= Math.Floor(xMax/xScale); i++)
				{
					if(i == 0)
					{
						continue;
					}
					float xi = (float)( w*(i*xScale-xWindowMin)/xWindowRange );
					float yi = (float)( h*(1.0-(-yWindowMin)/yWindowRange) );
					string str = (i*xScale).ToString();
					SizeF size = e.Graphics.MeasureString(str, f);
					e.Graphics.DrawString(str, f, Brushes.Black, xi - 0.5f*size.Width, yi + 3.0f);
					e.Graphics.DrawLine(Pens.Black, xi, yi, xi, yi + 3.0f);
				}

				//draw divisions along y axis
				for(int i = (int)Math.Ceiling(yMin/yScale); i <= Math.Floor(yMax/yScale); i++)
				{
					if(i == 0)
					{
						continue;
					}
					float xi = (float)( w*(-xWindowMin)/xWindowRange );
					float yi = (float)( h*(1.0-(i*yScale-yWindowMin)/yWindowRange) );
					string str = (i*yScale).ToString();
					SizeF size = e.Graphics.MeasureString(str, f);
					e.Graphics.DrawString(str, f, Brushes.Black, xi - size.Width - 3.0f, yi - 0.5f*size.Height);
					e.Graphics.DrawLine(Pens.Black, xi, yi, xi - 3.0f, yi);
				}

				// draw data
				PointF[] points = new PointF[d.Length];
				for(int i = 0; i < d.Length; i++)
				{
					points[i] = new PointF((float)( w*(d[i].X-xWindowMin)/xWindowRange ),
						(float)( h*(1.0-(d[i].Y-yWindowMin)/yWindowRange) ));
				}
				if(radius == 0)
				{
					e.Graphics.DrawLines(Pens.Blue, points);
				}
				else
				{
					int radiusTimes2 = 2*radius;
					foreach(PointF p in points)
					{
						e.Graphics.FillEllipse(Brushes.Blue, p.X-radius, p.Y-radius, radiusTimes2, radiusTimes2); 
					}
				}
				points = null;
			}
			catch(OverflowException)
			{}
		}

		protected override void OnResize(System.EventArgs e)
		{
			Invalidate(); // want to reorganize graph on resize
			base.OnResize(e);
		}
	}
}
